#ifndef CCLOTH_SIMULATION_H
#define CCLOTH_SIMULATION_H

#include "IAssignment.h"
#include "CTriMesh.h"
#include "texture.h"

#define NUM_BANKS	32

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CClothSimulation
	:public IAssignment
{
protected:

	unsigned int			m_ClothResX;
	unsigned int			m_ClothResY;
	unsigned int			m_FrameCounter;

	CTriMesh*				m_pClothModel;
	CTriMesh*				m_pEnvironment;
	OGLTexture*				m_pClothTexture;
	GLUquadricObj*			m_pSphere;
	float					m_SphereRadius;
	float4					m_SpherePos;

	// OpenGL variables
	GLhandleARB				m_VSCloth;
	GLhandleARB				m_PSCloth;
	GLhandleARB				m_ProgRenderCloth;

	GLhandleARB				m_VSMesh;
	GLhandleARB				m_PSMesh;
	GLhandleARB				m_ProgRenderMesh;

	// OpenCL variables

	//device buffers for cloth simulation
	cl_mem					m_clPosArray; //current position of the particles (shared with OpenGL)
	cl_mem					m_clPosArrayOld;
	cl_mem					m_clPosArrayAux;
	cl_mem					m_clNormalArray;

	cl_program				m_ClothSimProgram;
	cl_kernel				m_NormalKernel;
	cl_kernel				m_IntegrateKernel;
	cl_kernel				m_ConstraintKernel;
	cl_kernel				m_CollisionsKernel;

	float					m_ElapsedTime;
	float					m_PrevElapsedTime;

	bool					m_KeyboardMask[255];
	bool					m_InspectCloth;

	// mouse
	int						m_MouseState;
	int						m_Buttons;
	int						m_PrevX;
	int						m_PrevY;

	//for camera handling
	float					m_RotateX;
	float					m_RotateY;
	float					m_TranslateZ;

public:
	CClothSimulation(unsigned int ClothResX, unsigned int ClothResY);

	virtual ~CClothSimulation();
	
	virtual bool InitResources(cl_device_id Device, cl_context Context);
	
	virtual void ReleaseResources();

	virtual void ComputeGPU(cl_context Context, cl_command_queue CommandQueue, size_t LocalWorkSize[3]);

	virtual void Render();

	virtual void OnKeyboard(unsigned char Key, int X, int Y);

	virtual void OnKeyUp(unsigned char Key, int X, int Y);

	virtual void OnMouse(int Button, int State, int X, int Y);

	virtual void OnMouseMove(int X, int Y);

	virtual void OnIdle(double Time, float ElapsedTime);

	virtual void OnWindowResized(int Width, int Height);
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#endif
