/*
*/

#ifndef CSCAN_H
#define CSCAN_H

#include "IAssignment.h"
#include "CTriMesh.h"

#define NUM_BANKS	32

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class CParticleSystem
	:public IAssignment
{
protected:
	unsigned int		m_nParticles;
	unsigned int		m_nTriangles;
	unsigned int		m_volumeRes[3];

	CTriMesh*			m_pMesh;

	// OpenCL memory objects
	// arrays for particle data
	cl_mem				m_clPosLife[2];
	cl_mem				m_clVelMass[2];
	cl_mem				m_clAlive;
	cl_mem				m_clTriangleSoup;
	cl_mem				m_clRank;
	cl_mem				m_clVolTex3D;

	// ADD YOUR CODE HERE
	// YOU WILL NEED A PROGRAM, ARRAYS, AND KERNELS FOR THE PPS

	// OpenCL program and kernels
	cl_sampler			m_LinearSampler;
	cl_program			m_PSystemProgram;
	cl_kernel			m_IntegrateKernel;
	cl_kernel			m_ClearKernel;
	cl_kernel			m_ReorganizeKernel;

	// OpenGL variables
	//these will be used as VBOs
	GLuint				m_glPosLife[2];
	//these will be used as TBOs
	GLuint				m_glVelMass[2];
	GLuint				m_glVolTex3D;
	//texture objects (like the SRV in DirectX) for the TBOs
	GLuint				m_glTexVelMass[2];

	GLhandleARB			m_VSParticles;
	GLhandleARB			m_PSParticles;
	GLhandleARB			m_ProgRenderParticles;

	GLhandleARB			m_VSMesh;
	GLhandleARB			m_PSMesh;
	GLhandleARB			m_ProgRenderMesh;

	GLuint				m_glForceLines;
	GLhandleARB			m_VSForceField;
	GLhandleARB			m_PSForceField;
	GLhandleARB			m_ProgRenderForceField;
	
	size_t				m_LocalWorkSize[3];

	bool				m_KeyboardMask[255];
	bool				m_ShowForceField;

	// mouse
	int					m_MouseState;
	int					m_Buttons;
	int					m_PrevX;
	int					m_PrevY;

	//for camera handling
	float				m_RotateX;
	float				m_RotateY;
	float				m_TranslateZ;

	void Scan(cl_context Context, cl_command_queue CommandQueue, size_t LocalWorkSize[3]);
	void Integrate(cl_context Context, cl_command_queue CommandQueue, size_t LocalWorkSize[3], float dT);
	void Reorganize(cl_context Context, cl_command_queue CommandQueue, size_t LocalWorkSize[3]);

public:
	CParticleSystem(unsigned int N, size_t LocalWorkSize[3]);

	virtual ~CParticleSystem();
	
	virtual bool InitResources(cl_device_id Device, cl_context Context);
	
	virtual void ReleaseResources();

	virtual void ComputeGPU(cl_context Context, cl_command_queue CommandQueue, size_t LocalWorkSize[3]);

	virtual void Render();

	virtual void OnKeyboard(unsigned char Key, int X, int Y);

	virtual void OnKeyUp(unsigned char Key, int X, int Y);

	virtual void OnMouse(int Button, int State, int X, int Y);

	virtual void OnMouseMove(int X, int Y);

	virtual void OnIdle(double Time, float ElapsedTime);

	virtual void OnWindowResized(int Width, int Height);
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#endif