#define DAMPING 0.02

#define G_ACCEL (float4)(0.f, -9.81f, 0.f, 0.f)

#define WEIGHT_ORTHO	0.138f
#define WEIGHT_DIAG		0.097f
#define WEIGHT_ORTHO_2	0.069f
#define WEIGHT_DIAG_2	0.048f


#define ROOT_OF_2 1.4142135f
#define DOUBLE_ROOT_OF_2 2.8284271f




///////////////////////////////////////////////////////////////////////////////
__kernel void Integrate(unsigned int width, 
						unsigned int height, 
						__global float4* d_pos,
						__global float4* d_posOld,
						float elapsedTime,
						float prevElapsedTime){

	// Make sure the work-item does not map outside the cloth
    if(get_global_id(0) >= width || get_global_id(1) >= height)
		return;

	// This is just to keep every 8th particle of the first row attached to the bar
	unsigned int particleID = get_global_id(0) + get_global_id(1) * width;
    if(particleID > width-1 || ( particleID & ( 7 )) != 0){

		// ADD YOUR CODE HERE!
		// Read the old position
		// Compute the new one using the Verlet position integration
		// Write it back.

		// Hint: if the elapsedTime == 0.f (the first frame), set the previous velocity in the integration to 0.
    }
}



float4 SatisfyConstraint(float4 pos1,
						 float4 pos2,
						 float restDistance){
	float4 toNeighbor = pos2 - pos1;
	//float dist = length(toNeighbor);
	return (toNeighbor - normalize(toNeighbor) * restDistance);
}

///////////////////////////////////////////////////////////////////////////////
__kernel void SatisfyConstraints(unsigned int width, 
								unsigned int height, 
								float restDistance,
								__global float4* d_pos,
								__global float4* d_posOld){
    
    if(get_global_id(0) >= width || get_global_id(1) >= height)
		return;

	// ADD YOUR CODE HERE!
	// Satisfy all the constraints (structural, shear, and bend).
	// You can use weights defined at the beginning of this file.

}


///////////////////////////////////////////////////////////////////////////////

__kernel void CheckCollisions(	unsigned int width, 
								unsigned int height, 
								__global float4* d_pos,
								float4 spherePos,
								float sphereRad){
								
	// ADD YOUR CODE HERE!
	// Find whether the particle is inside the sphere.
	// If so, push it outside.
								
}

///////////////////////////////////////////////////////////////////////////////

float4 CalcTriangleNormal( float4 p1, float4 p2, float4 p3) {
    float4 v1 = p2-p1;
    float4 v2 = p3-p1;

    return cross( v1, v2);
}

__kernel void ComputeNormals(	unsigned int width, 
								unsigned int height, 
								__global float4* d_pos,
								__global float4* d_normal){
								
    int particleID = get_global_id(0) + get_global_id(1) * width;
    float4 normal = (float4)( 0.0f, 0.0f, 0.0f, 0.0f);
    
    int minX, maxX, minY, maxY, cntX, cntY;
    minX = max( (int)(0), (int)(get_global_id(0)-1));
    maxX = min( (int)(width-1), (int)(get_global_id(0)+1));
    minY = max( (int)(0), (int)(get_global_id(1)-1));
    maxY = min( (int)(height-1), (int)(get_global_id(1)+1));
    
    for( cntX = minX; cntX < maxX; ++cntX) {
        for( cntY = minY; cntY < maxY; ++cntY) {
            normal += normalize( CalcTriangleNormal(
                d_pos[(cntX+1)+width*(cntY)],
                d_pos[(cntX)+width*(cntY)],
                d_pos[(cntX)+width*(cntY+1)]));
            normal += normalize( CalcTriangleNormal(
                d_pos[(cntX+1)+width*(cntY+1)],
                d_pos[(cntX+1)+width*(cntY)],
                d_pos[(cntX)+width*(cntY+1)]));
        }
    }
    d_normal[particleID] = normalize( normal);
}
