
#ifndef _HLSL_EX_H_
#define _HLSL_EX_H_

#include "HLSL.h"
/******************************************************************************/
/*                                   Types                                    */
/******************************************************************************/

template <typename T>
struct quaternion {
	quaternion() {
	}

	quaternion(T x, T y, T z, T w) : x(x), y(y), z(z), w(w) {
	}

	T x, y, z, w;
};

template <typename T>
bool operator==(const quaternion<T> & a, const quaternion<T> & b) {
	return a.x == b.x && a.y == b.y && a.z == b.z && a.w == b.w;
}

template <typename T>
bool operator!=(const quaternion<T> & a, const quaternion<T> & b) {
	return a.x != b.x || a.y != b.y || a.z != b.z || a.w != b.w;
}

template <typename T>
quaternion<T> operator+(const quaternion<T> & a, const quaternion<T> & b) {
	return quaternion<T>(a.x + b.x, a.y + b.y, a.z + b.z, a.w + b.w);
}

template <typename T>
quaternion<T> operator-(const quaternion<T> & a, const quaternion<T> & b) {
	return quaternion<T>(a.x - b.x, a.y - b.y, a.z - b.z, a.w - b.w);
}

template <typename T>
quaternion<T> operator*(T a, const quaternion<T> & b) {
	return quaternion<T>(a * b.x, a * b.y, a * b.z, a * b.w);
}

template <typename T>
quaternion<T> operator*(const quaternion<T> & a, const quaternion<T> & b) {
	quaternion<T> c;

	c.x = a.w * b.x + b.w * a.x + a.y * b.z - a.z * b.y;
	c.y = a.w * b.y + b.w * a.y + a.z * b.x - a.x * b.z;
	c.z = a.w * b.z + b.w * a.z + a.x * b.y - a.y * b.x;
	c.w = a.w * b.w - a.x * b.x - a.y * b.y - a.z * b.z;

	return c;
}

/******************************************************************************/

//typedef unsigned __int8	 uchar;
typedef unsigned char uchar;

typedef quaternion<float> floatq;

/******************************************************************************/
/*                                   Macros                                   */
/******************************************************************************/

#define HLSL_EX_PI		3.141592653589793238462643383279502884197169399375105820974944592307816406286208998628034825f

/******************************************************************************/
/*                                 Functions                                  */
/******************************************************************************/

template <typename T, int M, int N>
const matrix<T, N, M> adjoint(const matrix<T, M, N> & a) {
	matrix<T, N, M> b;
	for (int i = 0; i < M; i++) {
		for (int j = 0; j < N; j++) {
			int s = 1 - 2 * ((i ^ j) & 1);
			b[j][i] = (T) s * cofactor(a, i, j);
		}
	}
	return b;
}

/*
template <typename T, int M, int N>
matrix<T, M, N> asmatrix(const quaternion<T> & a) {
	T _xx = a.x * a.x;
	T _xy = a.x * a.y;
	T _xz = a.x * a.z;
	T _yy = a.y * a.y;
	T _yz = a.y * a.z;
	T _zz = a.z * a.z;
	T _wx = a.w * a.x;
	T _wy = a.w * a.y;
	T _wz = a.w * a.z;

	matrix<T, M, N> b = identity<T, M, N>();

	b[0][0] = (T) 1 - (T) 2 * (_yy + _zz);
	b[0][1] = (T) 2 * (_xy - _wz);
	b[0][2] = (T) 2 * (_xz + _wy);

	b[1][0] = (T) 2 * (_xy + _wz);
	b[1][1] = (T) 1 - (T) 2 * (_xx + _zz);
	b[1][2] = (T) 2 * (_yz - _wx);

	b[2][0] = (T) 2 * (_xz - _wy);
	b[2][1] = (T) 2 * (_yz + _wx);
	b[2][2] = (T) 1 - (T) 2 * (_xx + _yy);

	return b;
}
*/

template <typename T, int M, int N>
T cofactor(const matrix<T, M, N> & a, int i, int j) {
	matrix<T, M - 1, N - 1> b;

	for (int jj = 0; jj < j; jj++) {
		for (int ii = 0; ii < i; ii++) {
			b[ii][jj] = a[ii][jj];
		}
	}

	for (int jj = j + 1; jj < N; jj++) {
		for (int ii = 0; ii < i; ii++) {
			b[ii][jj - 1] = a[ii][jj];
		}
	}

	for (int jj = 0; jj < j; jj++) {
		for (int ii = i + 1; ii < M; ii++) {
			b[ii - 1][jj] = a[ii][jj];
		}
	}

	for (int jj = j + 1; jj < N; jj++) {
		for (int ii = i + 1; ii < M; ii++) {
			b[ii - 1][jj - 1] = a[ii][jj];
		}
	}

	return determinant(b);
}

template <typename T>
T cofactor(const matrix<T, 2, 2> & a, int i, int j) {
	return a[1 - i][1 - j];
}

template <typename T>
quaternion<T> conjugate(const quaternion<T> & a) {
	return quaternion<T>(-a.x, -a.y, -a.z, a.w);
}

template <typename T, int N>
inline T distance_sqr(const vector<T, N> & a, const vector<T, N> & b) {
	return length_sqr(a - b);
}

template <typename T>
T dot(const quaternion<T> & a, const quaternion<T> & b) {
	return a.x * b.x + a.y * b.y + a.z * b.z + a.w * b.w;
}

template <typename T>
matrix<T, 4, 4> frustum(T x0, T x1, T y0, T y1, T z0, T z1) {
	return matrix<T, 4, 4>(
		(T) 2 * z0 / (x1 - x0),                  (T) 0,                (T) 0, (T) 0, 
		                 (T) 0, (T) 2 * z0 / (y1 - y0),                (T) 0, (T) 0, 
		 (x0 + x1) / (x0 - x1),  (y0 + y1) / (y0 - y1),       z1 / (z1 - z0), (T) 1, 
		                 (T) 0,                  (T) 0, -z1 * z0 / (z1 - z0), (T) 0
	);
}

template <typename T, int M, int N>
const matrix<T, M, N> identity() {
	matrix<T, M, N> a;
	for (int i = 0; i < M; i++) {
		for (int j = 0; j < N; j++) {
			a[i][j] = i == j ? (T) 1 : (T) 0;
		}
	}
	return a;
}

template <typename T, int N>
inline const matrix<T, N, N> invert(const matrix<T, N, N> & a) {
	return ((T) 1 / determinant(a)) * adjoint(a);
}

template <typename T>
quaternion<T> invert(const quaternion<T> & a) {
	T s = (T) 1 / length_sqr(a);
	return quaternion<T>(-s * a.x, -s * a.y, -s * a.z, s * a.w);
}

template <typename T, int N>
T length_sqr(const vector<T, N> & a) {
	return dot(a, a);
}

template <typename T>
T length_sqr(const quaternion<T> & a) {
	return dot(a, a);
}

template <typename T>
T length(const quaternion<T> & a) {
	return sqrt(length_sqr(a));
}

template <typename T>
quaternion<T> normalize(const quaternion<T> & a) {
	T s = (T) 1 / length(a);
	return quaternion<T>(a.x *s, a.y * s, a.z * s, a.w * s);
}

template <typename T>
matrix<T, 4, 4> lookat(const vector<T, 3> & eye, const vector<T, 3> & at, const vector<T, 3> & up) {
	vector<T, 3> zaxis = normalize(at - eye);
	vector<T, 3> xaxis = normalize(cross(up, zaxis));
	vector<T, 3> yaxis = cross(zaxis, xaxis);

	return matrix<T, 4, 4>(
		 xaxis.x,          yaxis.x,          zaxis.x,         (T) 0,
		 xaxis.y,          yaxis.y,          zaxis.y,         (T) 0,
		 xaxis.z,          yaxis.z,          zaxis.z,         (T) 0,
		-dot(xaxis, eye), -dot(yaxis, eye), -dot(zaxis, eye), (T) 1);
}

template <typename T>
vector<T, 3> mul(const vector<T, 3> & a, const quaternion<T> & b) {
	quaternion<T> q = invert(b) * quaternion<T>(a.x, a.y, a.z, 0.0f) * b;
	return vector<T, 3>(q.x, q.y, q.z);
}

template <typename T>
vector<T, 3> mul_unit(const vector<T, 3> & a, const /* unit */ quaternion<T> & b) {
	quaternion<T> q = conjugate(b) * quaternion<T>(a.x, a.y, a.z, 0.0f) * b;
	return vector<T, 3>(q.x, q.y, q.z);
}

template <typename T>
matrix<T, 4, 4> perspective(T fovy, T aspect, T z0, T z1) {
	T y1 = z0 * tan(fovy / (T) 2);
	T y0 = -y1;
	T x1 = aspect * y1;
	T x0 = -x1;
	return frustum<T>(x0, x1, y0, y1, z0, z1);
}

template<typename T>
matrix<T, 4, 4> ortho(T x0, T x1, T y0, T y1, T z0, T z1) {
	return matrix<T, 4, 4>(
		(T) 2 / (x1 - x0),     (T) 0,                 (T) 0,             (T) 0,
		(T) 0,                 (T) 2 / (y1 - y0),     (T) 0,             (T) 0,
		(T) 0,                 (T) 0,                 (T) 1 / (z1 - z0), (T) 0,
		(x0 + x1) / (x0 - x1), (y1 + y0) / (y0 - y1), z0 / (z0 - z1),    (T) 1);
}

template <typename T>
quaternion<T> rotation(const /* unit */ vector<T, 3> & axis, T angle) {
	T ca = cos(angle / (T) 2);
	T sa = sin(angle / (T) 2);

	return quaternion<T>(sa * axis.x, sa * axis.y, sa * axis.z, ca);
}

template <typename T, int M, int N>
matrix<T, M, N> rotation_x(T angle) {
	T ca = cos(angle);
	T sa = sin(angle);

	matrix<T, M, N> a = identity<T, M, N>();
	a[1][1] = +ca;
	a[1][2] = +sa;
	a[2][1] = -sa;
	a[2][2] = +ca;
	return a;
}

template <typename T, int M, int N>
matrix<T, M, N> rotation_y(T angle) {
	T ca = cos(angle);
	T sa = sin(angle);

	matrix<T, M, N> a = identity<T, M, N>();
	a[0][0] = +ca;
	a[0][2] = -sa;
	a[2][0] = +sa;
	a[2][2] = +ca;
	return a;
}

template <typename T, int M, int N>
matrix<T, M, N> rotation_z(T angle) {
	T ca = cos(angle);
	T sa = sin(angle);

	matrix<T, M, N> a = identity<T, M, N>();
	a[0][0] = +ca;
	a[0][1] = +sa;
	a[1][0] = -sa;
	a[1][1] = +ca;
	return a;
}

template <typename T, int M, int N>
matrix<T, M, N> scale(const vector<T, 3> & s) {
	matrix<T, M, N> a = identity<T, M, N>();
	for (int i = 0; i < 3; i++) {
		a[i][i] = s[i];
	}
	return a;
}

template <typename T>
matrix<T, 4, 4> scale4x4(const vector<T, 3>& s)
{
	return scale<T, 4, 4>(s);
}

template <typename T>
quaternion<T> slerp(const quaternion<T> & a, const quaternion<T> & b, T t) {
	T theta = acos(dot(a, b));
	T t0 = sin(((T) 1 - t) * theta);
	T t1 = sin(         t  * theta);
	T s = (T) 1 / sin(theta);

	return s * (t0 * a + t1 * b);
}

template <typename T>
inline T sqr(T a) {
	return a * a;
}

template <typename T, int M, int N>
matrix<T, 4, 4> translation(const vector<T, 3> & t) {
	matrix<T, M, N> a = identity<T, M, N>();
	for (int j = 0; j < 3; j++) {
		a[3][j] = t[j];
	}
	return a;
}

template <typename T>
matrix<T, 4, 4> translation4x4(const vector<T,3>& t)
{
	return translation<T, 4, 4>(t);
}

#endif
